'use strict';
const http = require('http');
const https = require('https');
const Http2ClientRequest = require('./client-request');
const httpResolveALPN = require('./utils/http-resolve-alpn');
const urlToOptions = require('./utils/url-to-options');

const cache = {};

const prepareRequest = async (input, options) => {
	if (typeof input === 'string' || input instanceof URL) {
		input = urlToOptions(new URL(input));
	}

	options = {
		protocol: 'https:',
		...input,
		...options
	};

	if (options.protocol === 'https:') {
		const host = options.hostname || options.host || 'localhost';
		const port = options.port || 443;
		const ALPNProtocols = options.ALPNProtocols || ['h2', 'http/1.1'];
		const name = `${host}:${port}:${ALPNProtocols.sort()}`;

		let alpnProtocol = cache[name];

		if (typeof alpnProtocol === 'undefined') {
			alpnProtocol = (await httpResolveALPN(options)).alpnProtocol;
			cache[name] = alpnProtocol;

			const keys = Object.keys(cache);

			while (keys.length > 100) {
				delete cache[keys.pop()];
			}
		}

		if (alpnProtocol === 'h2') {
			if (options.agent && options.agent.http2) {
				options.agent = options.agent.http2;
			}

			return {
				options,
				request: Http2ClientRequest.request
			};
		}

		if (options.agent && options.agent.https) {
			options.agent = options.agent.https;
		}

		options._defaultAgent = https.globalAgent;
	} else if (options.agent && options.agent.http) {
		options.agent = options.agent.http;
	}

	options.session = options.socketSession;

	return {
		options,
		request: http.request
	};
};

module.exports = async (input, options, callback) => {
	const {options: preparedOptions, request} = await prepareRequest(input, options);

	return request(preparedOptions, callback);
};

module.exports.resolveALPN = httpResolveALPN;
module.exports.prepareRequest = prepareRequest;
module.exports.protocolCache = cache;
